def bind(f, *args, **kwargs):
    return lambda a: f(*args, a, **kwargs)


def rbind(f, *args, **kwargs):
    return lambda a: f(a, *args, **kwargs)


def fst(t):
    return t[0]


def snd(t):
    return t[1]


def cmap(f):
    "curryed map"
    return lambda l: map(f, l)


def cfilter(f):
    "curryed filter"
    return lambda l: filter(f, l)


def pipe(v, *args):
    for f in args:
        v = f(v)
    return v


def compose(*args):
    return lambda v: pipe(v, *args)



from yst.hack.forbiddenfruit import curse


def size(self):
    return len(self)


def empty(self):
    return len(self) == 0


def _map(self, f):
    return map(f, self)


def _filter(self, f):
    return filter(f, self)


def dict_filter(self, f: dict):
    return filter(f, self.items())


def dict_map(self: dict, f):
    return map(f, self.items())


def to_list(self):
    return list(self)


def to_dict(self):
    return dict(self)


def to_set(self):
    return set(self)


for t in [map, filter, list, set, dict]:
    for (n, f) in [('to_list', to_list), ('to_dict', to_dict), ('to_set', to_set)]:
        curse(t, n, f)

for t in [list, set]:
    curse(t, 'map', _map)
    curse(t, 'filter', _filter)
curse(dict, 'map', dict_map)
curse(dict, 'filter', dict_filter)

for t in [str, list, map, set]:
    curse(t, "size", size)
    curse(t, "empty", empty)



from yst.fp import *
from operator import add, mul


def foo(a, b, c):
    return [a, b, c]


assert bind(foo, 1, c=3)(2) == [1, 2, 3]
assert rbind(foo, 1, c=3)(2) == [2, 1, 3]

assert fst((1, 2)), 1
assert snd((1, 2)), 2

add1 = bind(add, 1)
mul3 = bind(mul, 3)
add1mul3 = compose(add1, mul3)
assert pipe(1, add1mul3) == 6

assert pipe(range(10),
            cmap(mul3),
            cfilter(lambda x: x % 2 == 0),
            list) == [6, 12, 18, 24, 30]
